/*
**  Copyright (C) 1993/1994, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/

/*
** Image Server Central Test Node (ctn) Main Program
**
** Author: 	Andrew Hewett
**		Universitaet Oldenburg, Germany
** Created:	08/1993
**
** Last Update:		$Author: meichel $
** Update Date:		$Date: 1997/05/29 17:06:34 $
** Source File:		$Source: /share/dicom/cvs-depot/dcmtk/imagectn/apps/imagectn.cc,v $
** CVS/RCS Revision:	$Revision: 1.12 $
** Status:		$State: Exp $
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <time.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <errno.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_SYS_ERRNO_H
#include <sys/errno.h>
#endif
END_EXTERN_C

#include "dicom.h"
#include "imagectn.h"
#include "dimse.h"
#include "cnf.h"
#include "sce.h"
#ifndef NODISPLAY
#include "dispuser.h"
#endif
#include "imagedb.h"
#include "dcdict.h"
#include "dcdebug.h"
#include "cmdlnarg.h"
#include "dcuid.h"    /* for dcmtk version name */

static char rcsid[] = "$dcmtk: imagectn v"
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";

#define APPLICATIONTITLE	"IMAGECTN"

typedef struct _ctn_processslot CTN_ProcessSlot;

struct _ctn_processslot {
	DIC_NODENAME		peerName;
	DIC_AE			callingAETitle;
	DIC_AE			calledAETitle;
	int			processId;
	time_t			startTime;
	BOOLEAN			hasStorageAbility;
	CTN_ProcessSlot		*next;
};

typedef struct {
	int			pcnt;	
	CTN_ProcessSlot		*plist;
} CTN_ProcessTable;

static CTN_ProcessTable	processTable;

char *nonStorageSyntaxes[] = {
    UID_VerificationSOPClass, 
    UID_FINDPatientRootQueryRetrieveInformationModel, 
    UID_MOVEPatientRootQueryRetrieveInformationModel, 
    UID_FINDStudyRootQueryRetrieveInformationModel, 
    UID_MOVEStudyRootQueryRetrieveInformationModel
};

int numberOfNonStorageSyntaxes = DIM_OF(nonStorageSyntaxes);

/* Reasons for refusing association */
typedef enum {
	CTN_TooManyAssociations,
	CTN_CannotFork,
	CTN_BadAppContext,
	CTN_BadAEPeer,
	CTN_BadAEService,
	CTN_NoReason
} CTN_RefuseReason;


T_ASC_Network *net = NULL; /* the global DICOM network */

char* progname = NULL;
BOOLEAN verbose = FALSE;
BOOLEAN debug = FALSE;
BOOLEAN ignoreStoreData = FALSE;
static BOOLEAN singleProcess = FALSE;
static BOOLEAN rejectWhenNoImplementationClassUID = FALSE;
static BOOLEAN refuse = FALSE;
static int maxReceivePDULength = -1;
static char *configFileName = "configrc";
static BOOLEAN disableDisplay = TRUE;
BOOLEAN ctnDisplayCmdLineOverride = FALSE;
BOOLEAN requireFindForMove = FALSE;
BOOLEAN displayImagesOnMove = TRUE;
BOOLEAN displayTransferProgress = TRUE;
static BOOLEAN refuseMultipleStorageAssociations = FALSE;
static BOOLEAN checkFindIdentifier = FALSE;
static BOOLEAN checkMoveIdentifier = FALSE;
BOOLEAN restrictMoveToSameVendor = FALSE;
BOOLEAN restrictMoveToSameHost = FALSE;
BOOLEAN restrictMoveToSameAE = FALSE;

static int maxAssociations = 20;	/* initial guess */

static void
shortusage()
{
    fprintf(stderr, 
 "usage: %s [-c configrc][-D][-DD][-v][-d][-s][-r][-i][-b n][-M][-P][-F][-R]\n"
 "                [-u][-I peer][-XF][-XM][-ZA][-ZH][-ZV] [port]\n", 
 progname);
}

static void
fullusage()
{
    fprintf(stderr,"%s\n\n", rcsid);
    shortusage();
    fprintf(stderr, 
"parameters:\n"
"    port	tcp/ip port number to listen on (overrides value in \n"
"                config file)\n"
"options:\n"
"    -c configrc use specific configuration file (default: %s)\n"
"    -u          disable generation of unknown VR (UN)\n"
"    -D		enable display\n"
"    -DD		disable display\n"
"    -v		verbose mode\n"
"    -d		debug mode\n"
"    -s		single process\n"
"    -r		refuse association\n"
"    -i		ignore store data, receive but do not save to disk\n"
"    -I peer	enable display image on store/move for peer\n"
"    		can be specified multiple times, default: all enabled\n"
"    -M		do not display image on move\n"
"    -P		do not display image transfer progress\n"
"    -F		require FIND presentation context for MOVE\n"
"    -R		refuse multiple symultaneous STORE presentation contexts\n"
"    -XF		check C-FIND identifier validity\n"
"    -XM		check C-MOVE identifier validity\n"
"    -ZA		restrict move destination to requesting AE title\n"
"    -ZH		restrict move destination to requesting host\n"
"    -ZV		restrict move destination to requesting vendor\n"
"    -b n	set max receive PDU to n bytes\n",
    configFileName);
    exit(1);
}

static void 
usage()
{
    shortusage();
    exit(1);
}

void errmsg(const char* msg, ...)
{
	va_list args;
	
	fprintf(stderr, "%s: ", progname);
	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
}

static CONDITION waitForAssociation(T_ASC_Network *net);
static void cleanChildren();
static CONDITION negotiateAssociation(T_ASC_Association *assoc);
static void addPeerDisplayRights(const char* peerName);


int 
main(int argc, char *argv[])
{
    CONDITION cond;
    int i;
    int port = 0;
    int maxPDU;

#ifdef HAVE_GUSI_H
    /* needed for Macintosh */
    GUSISetup(GUSIwithSIOUXSockets);
    GUSISetup(GUSIwithInternetSockets);
#endif

    prepareCmdLineArgs(argc, argv, "imagectn");

    /* strip any leading path from program name */
    if ((progname = (char *)strrchr(argv[0], PATH_SEPARATOR)) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    if (argc < 1) {
	fullusage();
    }
    /* parse program arguments */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	switch (argv[i][1]) {
	case 'D':
	    disableDisplay = (argv[i][2] == 'D');
	    ctnDisplayCmdLineOverride = TRUE;
#ifdef NODISPLAY
	    fprintf(stderr, "Warning: not compiled for ctndisplay\n");
#endif	    
	    break;
	case 'u':
	    dcmEnableUnknownVRGeneration = TRUE;
	    break;
	case 'v':
	    verbose = TRUE;
	    break;
	case 'd':
	    debug = TRUE;
	    verbose = TRUE;
	    break;
	case 'c':
	    if (i++ < argc) {
		configFileName = argv[i];
	    } else
		usage();
	    break;
	case 's':
	    singleProcess = TRUE;
	    break;
	case 'r':
	    refuse = TRUE;
	    break;
	case 'i':
	    ignoreStoreData = TRUE;
	    break;
	case 'I':
	    if (i++ < argc) {
		addPeerDisplayRights(argv[i]);
	    } else
		usage();
	    break;
	case 'M':
	    displayImagesOnMove = FALSE;
	    break;
	case 'P':
	    displayTransferProgress = FALSE;
	    break;
	case 'F':
	    requireFindForMove = TRUE;
	    break;
	case 'R':
	    refuseMultipleStorageAssociations = TRUE;
	    break;
	case 'X':
	    switch (argv[i][2]) {
	    case 'F':
	        checkFindIdentifier = TRUE;
	        break;
	    case 'M':
	        checkMoveIdentifier = TRUE;
	        break;
	    default:
	        usage();
	        break;
	    }
	    break;
	case 'Z':
	    switch (argv[i][2]) {
	    case 'A':
		restrictMoveToSameAE = TRUE; break;
	    case 'H':
		restrictMoveToSameHost = TRUE; break;
	    case 'V':
		restrictMoveToSameVendor = TRUE; break;
	    default:
		usage(); break;
	    }
	    break;
	case 'b':
	    if (((i + 1) < argc) &&
		(sscanf(argv[i + 1], "%d", &maxReceivePDULength) == 1)) {
		i++;		/* Maximum Receive PDU Length parsed */
		if (maxReceivePDULength < ASC_MINIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too small",
			   maxReceivePDULength);
		    usage();
		} else if (maxReceivePDULength > ASC_MAXIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too big",
			   maxReceivePDULength);
		    usage();
		}
	    } else {
		usage();
	    }
	    break;
	default:
	    fullusage();
	}
    }

#ifdef USE_MALLOC_DEBUG
#ifdef HAVE_MALLOC_DEBUG
#define MALLOCDEBUGLEVEL 16
    if (verbose) {
        printf("malloc_debug on (%d)\n", MALLOCDEBUGLEVEL);
    }
    malloc_debug(MALLOCDEBUGLEVEL);
#endif
#endif

    if (access(configFileName, R_OK) < 0) {
	errmsg("cannot access %s: %s", configFileName, 
		strerror(errno));
	usage();
    }

    if (!CNF_init(configFileName)) {
	errmsg("bad config file: %s", configFileName);
	usage();
    }

    port = CNF_getNetworkTCPPort();
    if (port == 0) {
        /* not set, use default */
	port = 104;
    }

    if (i<argc) {
        /* get port number to listen on */
        if (sscanf(argv[i], "%d", &port) != 1) {
	    errmsg("bad port number: %s", argv[i]);
	    usage();
        }
    }
    
    maxPDU = CNF_getMaxPDUSize();
    if (maxPDU == 0) {
        /* not set, set default */
	maxPDU = ASC_DEFAULTMAXPDU;
    }
    
    if (maxPDU < ASC_MINIMUMPDUSIZE || maxPDU > ASC_MAXIMUMPDUSIZE) {
        errmsg("bad MaxPDUSize (%d) in config file: %s", 
	    maxPDU, configFileName);
	usage();
    }

    /* if nothing was set on command line then use config file value */
    if (maxReceivePDULength == -1) {
        maxReceivePDULength = maxPDU;
    }
    
    maxAssociations = CNF_getMaxAssociations();
        
    DUL_Debug(debug);
    DIMSE_debug(debug);
    DB_setDebugLevel(debug);	/* DB debugging */
    SetDebugLevel(((debug)?(3):(0)));	/* dcmdata debugging */
    
    DB_setIdentifierChecking(checkFindIdentifier, checkMoveIdentifier);

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded()) {
  	fprintf(stderr, "Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }
    
    processTable.pcnt = 0;
    processTable.plist = NULL;

    /* if port is privileged we must be as well */
    if (port < 1024) {
        if (geteuid() != 0) {
	    errmsg("cannot listen on port %d, insufficient privileges", port);
	    usage();
	}
    }
    
    cond = ASC_initializeNetwork(NET_ACCEPTORREQUESTOR, port, 10, &net);
    if (!SUCCESS(cond)) {
	errmsg("Error initialising network:");
	COND_DumpConditions();
	exit(1);
    }

    /* return to normal uid so that we can't do too much damage in case
     * things go very wrong.   Only relevant if the program is setuid root,
     * and run by another user.  Running as root user may be
     * potentially disasterous if this program screws up badly.
     */
    setuid(getuid());

    /* initialise display interface */
#ifndef NODISPLAY
    /* initialize display */
    if (!ctnDisplayCmdLineOverride) {
	disableDisplay = (!CNF_getDisplay());
    }
    DSU_setDisplayMode(!disableDisplay);
    if (!DSU_initDisplay(DSU_TIUser)) {
	disableDisplay = TRUE;
    }
#endif

    /* loop waiting for associations */
    while (SUCCESS(cond)) {

	cond = waitForAssociation(net);

	if (!singleProcess) {
	    cleanChildren();	/* clean up any child processes */
	}
    }

    cond = ASC_dropNetwork(&net);
    if (!SUCCESS(cond)) {
	errmsg("Error dropping network:");
	COND_DumpConditions();
	exit(1);
    }
    return 0;

}

/*
 * Process Table List Handling
 */

static void
addProcessToTable(int pid, T_ASC_Association * assoc)
{
    CTN_ProcessSlot	*ps;
    int i;

    if ((ps = (CTN_ProcessSlot*)malloc(sizeof(*ps))) == NULL) {
	errmsg("out of memory");
	return;
    }
    /* fill in info */
    ASC_getPresentationAddresses(assoc->params, ps->peerName, NULL);
    ASC_getAPTitles(assoc->params, ps->callingAETitle, 
	ps->calledAETitle, NULL);
    ps->processId = pid;
    ps->startTime = time(NULL);
    ps->hasStorageAbility = FALSE;
    for (i=0; i<numberOfDcmStorageSOPClassUIDs; i++) {
	if (ASC_findAcceptedPresentationContextID(
	    assoc, dcmStorageSOPClassUIDs[i])) {
	    ps->hasStorageAbility = TRUE;
	    break;	/* out of for loop */
	}
    }
    /* add to start of list */
    processTable.pcnt++;
    ps->next = processTable.plist;
    processTable.plist = ps;
}

#if 0
/* Unused */
static CTN_ProcessSlot* procWithCalledAETitle(const char* aeTitle)
{
    CTN_ProcessSlot	*ps;
    BOOLEAN		found = FALSE;

    if (aeTitle == NULL) return NULL;

    ps = processTable.plist;
    while (ps && !found) {
	found = strcmp(ps->calledAETitle, aeTitle) == 0;
	if (!found) ps = ps->next;
    }
    return ps;
}

static CTN_ProcessSlot* procWithCallingAETitle(const char* aeTitle)
{
    CTN_ProcessSlot	*ps;
    BOOLEAN		found = FALSE;

    if (aeTitle == NULL) return NULL;

    ps = processTable.plist;
    while (ps && !found) {
	found = strcmp(ps->callingAETitle, aeTitle) == 0;
	if (!found) ps = ps->next;
    }
    return ps;
}
#endif

#if 0
/* Unused */
static int
countProcessesForCalledAE(const char* aeTitle)
{
    int n = 0;
    CTN_ProcessSlot	*ps;

    if (aeTitle == NULL) return 0;

    ps = processTable.plist;
    while (ps) {
	if (strcmp(ps->calledAETitle, aeTitle) == 0) {
	    n++;
	}
	ps = ps->next;
    }
    return n;
}
#endif

#if 0
/* Unused */
static CTN_ProcessSlot* procWithPeerName(const char* peerName)
{
    CTN_ProcessSlot	*ps;
    BOOLEAN		found = FALSE;

    if (peerName == NULL) return NULL;

    ps = processTable.plist;
    while (ps && !found) {
	found = strcmp(ps->peerName, peerName) == 0;
	if (!found) ps = ps->next;
    }
    return ps;
}
#endif

static int
countChildProcesses()
{
    int n = 0;
    CTN_ProcessSlot	*ps;

    ps = processTable.plist;
    while (ps) {
	n++;
	ps = ps->next;
    }
    return n;
}

static void
removeProcessFromTable(int pid)
{
    CTN_ProcessSlot	*ps, *pps;
    BOOLEAN		found = FALSE;

    pps = NULL;
    ps = processTable.plist;

    while (ps && !found) {
	found = (ps->processId == pid);
	if (!found) {
	    pps = ps;
	    ps = ps->next;
	}
    }

    if (!found) {
	errmsg("WARNING: process %d not found", pid);
	return;
    }

    if (pps == NULL) {
	/* first in list */
	processTable.plist = ps->next;
    } else {
	pps->next = ps->next;
    }

    free(ps);
    processTable.pcnt--;
}

#ifdef HAVE_WAITPID
static void 
cleanChildren()
{
    int                 options;
    int			stat_loc;
    int                 child = 1;

    options = WNOHANG;
    while (child > 0) {
	child = waitpid(-1, &stat_loc, options);
	if (child == 0) {
	    /* child not yet finished so don't complain */
	} else if (child < 0) {
	    if (errno == ECHILD) {
		/* no children so don't complain */
	    } else {
		errmsg("wait for child failed: %s", strerror(errno));
	    }
	} else if (child > 0) {
	    if (verbose) {
		time_t t = time(NULL);
		printf("Cleaned up after child (%d) %s", child, ctime(&t));
	    }
	/* Remove Entry from Process Table */
	    removeProcessFromTable(child);
	}
    }
}
#elif HAVE_WAIT3
static void 
cleanChildren()
{
#if defined(__NeXT__)
    /* some systems need a union wait as argument to wait3 */
    union wait          status;
#else
    int                 status;
#endif
    int                 options;
    struct rusage       rusage;
    int                 child = 1;

    options = WNOHANG;
    while (child > 0) {
	child = wait3(&status, options, &rusage);
	if (child < 0) {
	    if (errno == ECHILD) {
		/* no children so don't complain */
	    } else {
		errmsg("wait for child failed: %s", strerror(errno));
	    }
	} else if (child > 0) {
	    if (verbose) {
		printf("Cleaned up after child (%d)\n", child);
	    }
	/* Remove Entry from Process Table */
	    removeProcessFromTable(child);
	}
    }
}
#else
/* don't know how to wait for my children */
static void 
cleanChildren()
{
    errmsg("cannot wait for child processes");
}
#endif


/*
 * Refuse an Association
 */

static CONDITION
refuseAssociation(T_ASC_Association ** assoc, CTN_RefuseReason reason)
{
    CONDITION           cond;
    T_ASC_RejectParameters rej;

    if (verbose) {
	printf("Refusing Association (");
	switch (reason) {
	case CTN_TooManyAssociations:
	    printf("TooManyAssociations");
	    break;
	case CTN_CannotFork:
	    printf("CannotFork");
	    break;
	case CTN_BadAppContext:
	    printf("BadAppContext");
	    break;
	case CTN_BadAEPeer:
	    printf("BadAEPeer");
	    break;
	case CTN_BadAEService:
	    printf("BadAEService");
	    break;
	case CTN_NoReason:
	    printf("NoReason");
	    break;
	default:
	    printf("???");
	    break;
	}
	printf(")\n");
    }
    switch (reason) {
    case CTN_TooManyAssociations:
	rej.result = ASC_RESULT_REJECTEDTRANSIENT;
	rej.source = ASC_SOURCE_SERVICEPROVIDER_PRESENTATION_RELATED;
	rej.reason = ASC_REASON_SP_PRES_LOCALLIMITEXCEEDED;
	break;
    case CTN_CannotFork:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEPROVIDER_PRESENTATION_RELATED;
	rej.reason = ASC_REASON_SP_PRES_TEMPORARYCONGESTION;
	break;
    case CTN_BadAppContext:
	rej.result = ASC_RESULT_REJECTEDTRANSIENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_APPCONTEXTNAMENOTSUPPORTED;
	break;
    case CTN_BadAEPeer:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_CALLINGAETITLENOTRECOGNIZED;
	break;
    case CTN_BadAEService:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_CALLEDAETITLENOTRECOGNIZED;
	break;
    case CTN_NoReason:
    default:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_NOREASON;
	break;
    }

    cond = ASC_rejectAssociation(*assoc, &rej);
    if (cond != ASC_NORMAL) {
	fprintf(stderr, "Association Reject Failed:\n");
	COND_DumpConditions();
    }

    cond = ASC_dropAssociation(*assoc);
    if (!SUCCESS(cond)) {
	fprintf(stderr, "Cannot Drop Association:\n");
	COND_DumpConditions();
    }
    cond = ASC_destroyAssociation(assoc);
    if (!SUCCESS(cond)) {
	fprintf(stderr, "Cannot Destroy Association:\n");
	COND_DumpConditions();
    }

    return cond;
}

/*
 * Wait for and negotiate incoming association.
 */

static CONDITION
waitForAssociation(T_ASC_Network * net)
{
    CONDITION           cond;
    int                 pid;
    T_ASC_Association  *assoc;
    char                buf[BUFSIZ];
    int timeout;

    if (singleProcess) {
	timeout = 1000;
    } else {
	if (countChildProcesses() > 0) {
	    timeout = 1;
	} else {
	    timeout = 1000;
	}
    }
    if (ASC_associationWaiting(net, timeout)) {
        cond = ASC_receiveAssociation(net, &assoc, maxReceivePDULength);
        if (!SUCCESS(cond)) {
	    if (verbose) {
	        errmsg("Failed to receive association:");
	        COND_DumpConditions();
	    }
	    goto cleanup;
        }
    } else {
	return ASC_NORMAL;
    }

    if (verbose) {
	time_t t = time(NULL);
	printf("Association Received (%s:%s -> %s) %s",
	       assoc->params->DULparams.callingPresentationAddress,
	       assoc->params->DULparams.callingAPTitle,
	       assoc->params->DULparams.calledAPTitle,
	       ctime(&t));
    }
    if (debug) {
	printf("Parameters:\n");
	ASC_dumpParameters(assoc->params);
    }
    if (refuse) {
	if (verbose) {
	    printf("Refusing Association (forced via command line)\n");
	}
	cond = refuseAssociation(&assoc, CTN_NoReason);
	goto cleanup;
    }

    /* Application Context Name */
    if ((cond = ASC_getApplicationContextName(assoc->params, buf) !=
	 ASC_NORMAL) || strcmp(buf, DICOM_STDAPPLICATIONCONTEXT) != 0) {
        /* reject: the application context name is not supported */
	if (verbose) {
	    errmsg("Bad AppContextName: %s", buf);
	}
	cond = refuseAssociation(&assoc, CTN_BadAppContext);
	goto cleanup;
    }

    /* Implementation Class UID */
    if (rejectWhenNoImplementationClassUID &&
	strlen(assoc->params->theirImplementationClassUID) == 0) {
        /* reject: no implementation Class UID provided */
	if (verbose) {
	    errmsg("No implementation Class UID provided");
	}
	cond = refuseAssociation(&assoc, CTN_NoReason);
	goto cleanup;
    }

    /* Does peer AE have access to required service ?? */
    if (! CNF_peerInAETitle(assoc->params->DULparams.calledAPTitle,
	assoc->params->DULparams.callingAPTitle,
	assoc->params->DULparams.callingPresentationAddress)) {			   
	cond = refuseAssociation(&assoc, CTN_BadAEService);
	goto cleanup;
    }

    /* 
     * too many concurrent associations ?? 
     */
    if (countChildProcesses() >= maxAssociations) {
	cond = refuseAssociation(&assoc, CTN_TooManyAssociations);
	goto cleanup;
    }

    cond = negotiateAssociation(assoc);
    if (cond != ASC_NORMAL) {
	goto cleanup;
    }

    cond = ASC_acknowledgeAssociation(assoc);
    if (cond != ASC_NORMAL) {
	COND_DumpConditions();
	goto cleanup;
    }
    if (verbose) {
	printf("Association Acknowledged (Max Send PDV: %lu)\n",
	       assoc->sendPDVLength);
	if (ASC_countAcceptedPresentationContexts(assoc->params) == 0)
	    printf("    (but no valid presentation contexts)\n");
	if (debug)
	    ASC_dumpParameters(assoc->params);
    }


    if (singleProcess) {
        /* don't spawn a sub-process to handle the association */
	cond = SCE_handleAssociation(assoc);
    } else {
        /* spawn a sub-process to handle the association */
	pid = fork();
	if (pid < 0) {
	    errmsg("Cannot create association sub-process: %s",
		   strerror(errno));
	    cond = refuseAssociation(&assoc, CTN_CannotFork);
	    goto cleanup;
	} else if (pid > 0) {
	    /* parent process, note process in table */

	    addProcessToTable(pid, assoc);

	} else {
	    /* child process, handle the association */
	    cond = SCE_handleAssociation(assoc);
	    /* the child process is done so exit */
	    exit(0);
	}

    }

cleanup:
    if (!singleProcess) {
        /* the child will handle the association, we can drop it */
        cond = ASC_dropAssociation(assoc);
        if (!SUCCESS(cond)) {
	   errmsg("Cannot Drop Association:");
	   COND_DumpConditions();
        }
        cond = ASC_destroyAssociation(&assoc);
        if (!SUCCESS(cond)) {
	    errmsg("Cannot Destroy Association:");
	    COND_DumpConditions();
        }
    }

    if (verbose) {
        /* are there any conditions sitting on the condition stack? */
	char                buf[BUFSIZ];
	CONDITION           c;

	if (COND_TopCondition(&c, buf, BUFSIZ) != COND_NORMAL) {
	    printf("CONDITIONS Remaining:");
	    COND_DumpConditions();
	}
    }
    COND_PopCondition(TRUE);	/* clean out condition stack */
    return cond;
}

static void
refuseAnyStorageContexts(T_ASC_Association * assoc)
{
    int i;
    T_ASC_PresentationContextID pid;

    for (i = 0; i < numberOfDcmStorageSOPClassUIDs; i++) {
	pid = ASC_findAcceptedPresentationContextID(assoc,
						    dcmStorageSOPClassUIDs[i]);
	if (pid != 0) {
	    /* refuse */
	    ASC_refusePresentationContext(assoc->params,
					  pid, ASC_P_USERREJECTION);
	}
    }
}

static BOOLEAN
writableStorageArea(const char *aeTitle)
{
    const char *access;
    access = CNF_getAccess((char*)aeTitle);
    if (strcmp(access, "RW") == 0) return TRUE;
    if (strcmp(access, "WR") == 0) return TRUE;
    if (strcmp(access, "W") == 0) return TRUE;
    return FALSE;
}

#if 0
/* Unused */
static BOOLEAN
readableStorageArea(const char *aeTitle)
{
    const char *access;
    access = CNF_getAccess((char*)aeTitle);
    if (strcmp(access, "RW") == 0) return TRUE;
    if (strcmp(access, "WR") == 0) return TRUE;
    if (strcmp(access, "R") == 0) return TRUE;
    return FALSE;
}
#endif

static CONDITION
negotiateAssociation(T_ASC_Association * assoc)
{
    CONDITION cond = ASC_NORMAL;
    int i;
    T_ASC_PresentationContextID movepid, findpid;
    struct { char *moveSyntax, *findSyntax; } queryRetrievePairs[] = {
	{ UID_MOVEPatientRootQueryRetrieveInformationModel, 
	  UID_FINDPatientRootQueryRetrieveInformationModel },
	{ UID_MOVEStudyRootQueryRetrieveInformationModel, 
	  UID_FINDStudyRootQueryRetrieveInformationModel },
	{ UID_MOVEPatientStudyOnlyQueryRetrieveInformationModel, 
	  UID_FINDPatientStudyOnlyQueryRetrieveInformationModel}
    };
    CTN_ProcessSlot *ps;
    DIC_AE calledAETitle;
    const char* transferSyntaxes[] = { 
	NULL, NULL, UID_LittleEndianImplicitTransferSyntax };

    ASC_getAPTitles(assoc->params, NULL, calledAETitle, NULL);

    /* 
    ** We prefer to accept Explicitly encoded transfer syntaxes.
    ** If we are running on a Little Endian machine we prefer 
    ** LittleEndianExplicitTransferSyntax to BigEndianTransferSyntax.
    */

    /* gLocalByteOrder is defined in dcxfer.h */
    if (gLocalByteOrder == EBO_LittleEndian) {
	/* we are on a little endian machine */
	transferSyntaxes[0] = UID_LittleEndianExplicitTransferSyntax;
	transferSyntaxes[1] = UID_BigEndianExplicitTransferSyntax;
    } else {
	/* we are on a big endian machine */
	transferSyntaxes[0] = UID_BigEndianExplicitTransferSyntax;
	transferSyntaxes[1] = UID_LittleEndianExplicitTransferSyntax;
    }

    /*  accept any of the non-storage syntaxes */
    cond = ASC_acceptContextsWithPreferredTransferSyntaxes(
	assoc->params, 
	(const char**)nonStorageSyntaxes, numberOfNonStorageSyntaxes,
	(const char**)transferSyntaxes, DIM_OF(transferSyntaxes));
    if (!SUCCESS(cond)) {
	errmsg("Cannot accept presentation contexts:");
	COND_DumpConditions();
    }

    /*  accept any of the storage syntaxes */
    cond = ASC_acceptContextsWithPreferredTransferSyntaxes(
	assoc->params, 
	dcmStorageSOPClassUIDs, numberOfDcmStorageSOPClassUIDs,
	(const char**)transferSyntaxes, DIM_OF(transferSyntaxes));
    if (!SUCCESS(cond)) {
	errmsg("Cannot accept presentation contexts:");
	COND_DumpConditions();
    }

    /*
     * Refuse any "Storage" presentation contexts to non-writable
     * storage areas.
     */
    if (!writableStorageArea(calledAETitle)) {
	refuseAnyStorageContexts(assoc);
    }

    /*
     * Enforce RSNA'93 Demonstration Requirements about only
     * accepting a context for MOVE if a context for FIND is also present. 
     */
    
    for (i=0; i<(int)DIM_OF(queryRetrievePairs); i++) {
        movepid = ASC_findAcceptedPresentationContextID(assoc, 
	    queryRetrievePairs[i].moveSyntax);
        if (movepid != 0) {
	    findpid = ASC_findAcceptedPresentationContextID(assoc, 
	        queryRetrievePairs[i].findSyntax);
	    if (findpid == 0) {
		if (requireFindForMove) {
		    /* refuse the move */
		    ASC_refusePresentationContext(assoc->params,
		        movepid, ASC_P_USERREJECTION);
	        } else {
		    errmsg("WARNING: Move PresCtx but no Find (accepting for now)");
		}
	    }
        }
    }

    /*
     * Enforce an Ad-Hoc rule to limit storage access.   
     * If the storage area is "writable" and some other association has
     * already negotiated a "Storage" class presentation context,
     * then refuse any "storage" presentation contexts.
     */

    if (refuseMultipleStorageAssociations) {
        if (writableStorageArea(calledAETitle)) {
	    ps = processTable.plist;
	    while (ps != NULL) {
	        if (strcmp(calledAETitle, ps->calledAETitle) == 0 &&
		    ps->hasStorageAbility) {
		    refuseAnyStorageContexts(assoc);
		    break;
	        }
	        ps = ps->next;
	    }
        }
    }

    return cond;
}


/*
 * A last minute hack to restrict the displaying images on store 
 * and move operations to a set of selected peers.  
 * i.e. those in infoRAD.
 * If no peers are specified then anyone can display, otherwise
 * only the specified peers.
 * This stuff should really go in the configuration file.
 */

#define MAX_DISPLAY_RIGHTS 100
static char* displayRights[MAX_DISPLAY_RIGHTS]; 
static int displayRightsCount = 0;

static void addPeerDisplayRights(const char* peerName)
{
    if (displayRightsCount == MAX_DISPLAY_RIGHTS) {
         errmsg("display rights array overflow");
    } else {
        displayRights[displayRightsCount] = 
	    (char*)strcpy((char*)malloc(strlen(peerName)+1), (char*)peerName);
        displayRightsCount++;
    }
}

BOOLEAN peerCanDisplayImages(const char* peerName)
{
    int i;
    
    if (displayRightsCount == 0) {
        return TRUE;
    }

    for (i=0; i<displayRightsCount; i++) {
        if (strcmp(peerName, displayRights[i]) == 0) {
	    return TRUE;
	}
    }
    
    return FALSE;
}


/*
** CVS Log
** $Log: imagectn.cc,v $
** Revision 1.12  1997/05/29 17:06:34  meichel
** All dcmtk applications now contain a version string
** which is displayed with the command line options ("usage" message)
** and which can be queried in the binary with the "ident" command.
**
** Revision 1.11  1997/05/22 13:31:39  hewett
** Modified the test for presence of a data dictionary to use the
** method DcmDataDictionary::isDictionaryLoaded().
**
** Revision 1.10  1997/03/27 16:11:39  hewett
** Added command line switches allowing generation of UN to
** be disabled (it is enabled by default).
**
** Revision 1.9  1997/02/06 12:24:45  hewett
** Added code to print the time when an association request
** is received.
**
** Revision 1.8  1996/09/27 08:46:19  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C.
**
** Revision 1.7  1996/09/24 15:52:34  hewett
** Now uses global table of Storage SOP Class UIDs (from dcuid.h).
** Also added preliminary support for the Macintosh environment (GUSI library).
**
** Revision 1.6  1996/05/30 17:41:26  hewett
** Added command line argument (-DD) to explicitly disable communication
** with ctndisp and override the configuration file.
**
** Revision 1.5  1996/05/06 07:33:13  hewett
** Rearranged handing of display disabling code.
**
** Revision 1.4  1996/04/25 16:28:09  hewett
** Reorded condition compiling for cleanChildren() function.  Code now
** prefers to use waitpid() rather than wait3() if available.
**
** Revision 1.3  1996/04/22 16:33:02  hewett
** Removed calls to DSU_setTraceLevel().
**
** Revision 1.2  1996/04/22 10:19:33  hewett
** Added command line flags to restrict move destination.  Added check for
** presence of a loaded data dictionary.  A global network is now created
** in acceptor/requestor mode.
**
** Revision 1.1.1.1  1996/03/28 19:24:59  hewett
** Oldenburg Image CTN Software ported to use the dcmdata C++ toolkit.
**
**
*/
