/*
**  Copyright (C) 1993/1994, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/
/*
** TI Common Constants, Types, Globals and Functions
**
** Author: 	Andrew Hewett
**		Universitaet Oldenburg, Germany
** Created:	08/1993
**
** Last Update:		$Author: hewett $
** Update Date:		$Date: 1996/04/22 10:27:25 $
** Source File:		$Source: /share/dicom/cvs-depot/dcmtk/imagectn/apps/ti.h,v $
** CVS/RCS Revision:	$Revision: 1.1 $
** Status:		$State: Exp $
**
** CVS/RCS Log at end of file
*/

#ifndef TI_H
#define TI_H

#include "osconfig.h"    /* make sure OS specific configuration is included first */

#include "dicom.h"
#include "cond.h"
#include "assoc.h"
#include "dimse.h"
#include "cnf.h"
#include "imagedb.h"

/*
 * Application Conditions
 */

#ifdef USEAPPCONDITIONS
#define APP_NORMAL		/* normal operation */\
	FORM_COND(FAC_APP, SEV_SUCC, 0)
#define APP_BADIMAGE		/* bad image data */\
	FORM_COND(FAC_APP, SEV_ERROR, 1)
#define APP_NOCONTEXT		/* no presentation context for SOP Class */\
	FORM_COND(FAC_APP, SEV_ERROR, 2)
#define APP_PROTOCOLERROR	/* protocol error */\
	FORM_COND(FAC_APP, SEV_ERROR, 3)
#define APP_MESSAGEFAILED	/* network message failed */\
	FORM_COND(FAC_APP, SEV_ERROR, 4)
#define APP_STOREFAILED		/* store operation failed */\
	FORM_COND(FAC_APP, SEV_ERROR, 5)
#endif

/*
 * Constants
 */

#define TI_MAXPEERS		100

#define TI_MAXDATABASES 	100
#define TI_MAXSTUDIES		1000
#define TI_MAXSERIES		500
#define TI_MAXIMAGES		1000

/*
 * Type definitions
 */

typedef struct {
    DIC_UI	sopInstanceUID;
    DIC_IS	imageNumber;
    int		intImageNumber;
} TI_ImageEntry;

typedef struct {
    DIC_UI	seriesInstanceUID;
    DIC_IS	seriesNumber;
    int		intSeriesNumber;
    DIC_CS	modality;
    TI_ImageEntry	*images[TI_MAXIMAGES];	/* array of image pointers */
    int			imageCount;

    time_t lastQueryTime;	/* time we last queried db */
} TI_SeriesEntry;

typedef struct {
    DIC_UI	studyInstanceUID;
    DIC_CS	studyID;
    DIC_PN	patientsName;
    DIC_LO	patientID;
    TI_SeriesEntry	*series[TI_MAXSERIES];	/* array of series pointers */
    int 		seriesCount;
    time_t lastQueryTime;	/* time we last queried db */
} TI_StudyEntry;


typedef struct {
    char *title;	/* the CTN AE Title associated with this DB */

    char **peerTitles;	/* peer titles which can read this database 
			 * and thus we can comminicate with */
    int peerTitleCount;	/* number of peer titles */

    DB_Handle *dbHandle;	/* handle to current db */

    TI_StudyEntry	*studies[TI_MAXSTUDIES]; /* array of study pointers */
    int			studyCount;

    int currentStudy;	/* index of current study */
    int currentSeries;  /* index of current series in current study */
    int currentImage;	/* index of current image in current study */

    time_t lastQueryTime;	/* time we last queried db */

    BOOL isRemoteDB;	/* true if DB is remote */ 

} TI_DBEntry;

typedef struct {
    TI_DBEntry **dbEntries;	/* the CTN databases we know */
    int dbCount;		/* number of entries in databases we know */

    char *peerHostName;		/* peer to talk to */
    char *peerNames[TI_MAXPEERS];
    int peerNamesCount;
    
    char *myAETitle;		/* my application entity title */

    T_ASC_Network *net;		/* active network */
    T_ASC_Association *assoc;	/* currently active association */

    int maxReceivePDULength;	/* number of bytes we can receive */

    int currentdb;		/* current database index */
    char *currentPeerTitle;	/* current peer title */

} TI_Config;


/*
 * Common Globals (defined in ti.c)
 */

extern char* progname;
extern BOOLEAN verbose;
extern BOOLEAN debug;
extern BOOLEAN disableDisplay;
extern BOOLEAN displayTransferProgress;
extern BOOLEAN disableImageDisplay;


/*
 * Common Function Definitions
 */

extern void errmsg(const char* msg, ...);



#endif

/*
** CVS Log
** $Log: ti.h,v $
** Revision 1.1  1996/04/22 10:27:25  hewett
** Initial release.
**
**
*/
