/*
**
**  Copyright (C) 1996, OFFIS
**
**  This software and supporting documentation were developed by
**
**    Kuratorium OFFIS e.V.
**    Forschungsbereich 2: Kommunikationssysteme
**    Escherweg 2
**    D-26121 Oldenburg, Germany
**
**  for CEN/TC251/WG4 as a contribution to the Computer Assisted Radiology
**  (CAR) 1996 DICOM Demonstration.
**
**  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
**  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
**  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
**  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
**
**  Copyright of the software  and  supporting  documentation  is,  unless
**  otherwise stated, owned by OFFIS, and free access is hereby granted as
**  a license to  use  this  software,  copy  this  software  and  prepare
**  derivative works based upon this software.  However, any  distribution
**  of this software source code or supporting documentation or derivative
**  works  (source code and  supporting documentation)  must  include  the
**  three paragraphs of this copyright notice.
**
*/
/*
** Service Class Executive (SCE)
**
** Author: 	Andrew Hewett
**		Kuratorium OFFIS e.V., Oldenburg, Germany
** Created:	03/1996
**
** Last Update:		$Author: hewett $
** Update Date:		$Date: 1996/09/27 08:54:25 $
** Source File:		$Source: /share/dicom/cvs-depot/dcmtk/wlistctn/apps/sce.cc,v $
** CVS/RCS Revision:	$Revision: 1.3 $
** Status:		$State: Exp $
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
END_EXTERN_C

#include "dicom.h"
#include "wlistctn.h"
#include "wrklstdb.h"
#include "dimse.h"
#include "diutil.h"
#include "sce.h"
#ifndef NODISPLAY
#include "dispuser.h"
#endif

#include "sceecho.h"
#include "scefind.h"


/*
 *
 */

static CONDITION
SCE_dispatch(T_ASC_Association *assoc)
{
    CONDITION cond = DIMSE_NORMAL;
    T_DIMSE_Message msg;
    T_ASC_PresentationContextID presID;
    char *myAE = NULL;
    
    myAE = assoc->params->DULparams.calledAPTitle;
        
    /* Create a database handle for this association */
    char wlistpath[2048];
    sprintf(wlistpath, "%s/%s", baseWorklistDBPath, 
	    assoc->params->DULparams.calledAPTitle);

    WorklistDB dbHandle(wlistpath);
 
    while (cond == DIMSE_NORMAL) {
	
	cond = DIMSE_receiveCommand(assoc, DIMSE_BLOCKING, 0, &presID,
		&msg, NULL);

	/* did peer release, abort, or do we have a valid message ? */
	switch (cond) {
	case DIMSE_NORMAL:
	    /* process command */
	    switch (msg.CommandField) {
	    case DIMSE_C_ECHO_RQ:
		cond = SCE_echoSCP(assoc, &msg.msg.CEchoRQ, presID);
		break;
	    case DIMSE_C_FIND_RQ:
#ifndef NODISPLAY
		DSU_operationState(DSU_MainAssociation, DSU_FindOperation,
			DSU_StartOperation, DSU_Incoming, 0);
#endif
		cond = SCE_findSCP(assoc, &msg.msg.CFindRQ, presID, 
		    dbHandle);
#ifndef NODISPLAY
		DSU_operationState(DSU_MainAssociation, DSU_FindOperation,
			DSU_StopOperation, DSU_Incoming, 100);
#endif
		break;
	    case DIMSE_C_CANCEL_RQ:
		/* This is a late cancel request, just ignore it */
		if (verbose) {
		    printf("SCE_dispatch: late C-CANCEL-RQ, ignoring\n");
		}
		break;
	    default:
		/* we cannot handle this kind of message */
		cond = DIMSE_BADCOMMANDTYPE;
		errmsg("Cannot handle command: 0x%x\n", 
			(unsigned)msg.CommandField);
	        /* the condition will be returned, the caller
	         * will abort the assosiation.
	         */
	    }
	    break;
	case DIMSE_PEERREQUESTEDRELEASE:
	case DIMSE_PEERABORTEDASSOCIATION:
	    /* association gone */
	    break;
	default:
	    /* the condition will be returned, the caller
	     * will abort the assosiation.
	     */
	    break;
	}
    }

    return cond;
}

CONDITION
SCE_handleAssociation(T_ASC_Association * assoc)
{
    CONDITION           cond;
    DIC_NODENAME	peerHostName;
    DIC_AE		peerAETitle;
    DIC_AE		myAETitle;

    ASC_getPresentationAddresses(assoc->params, peerHostName, NULL);
    ASC_getAPTitles(assoc->params, peerAETitle, myAETitle, NULL);

#ifndef NODISPLAY
    DSU_startAssociation(peerHostName, peerAETitle, myAETitle,
			 DSU_Acceptor, DSU_MainAssociation);
#endif
    
 /* now do the real work */
    cond = SCE_dispatch(assoc);

 /* clean up on association termination */
    if (cond == DIMSE_PEERREQUESTEDRELEASE) {
	COND_PopCondition(FALSE);	/* pop only the peer requested
					 * release condition from the stack */
	if (verbose)
	    printf("Association Release\n");
	cond = ASC_acknowledgeRelease(assoc);
    } else if (cond == DIMSE_PEERABORTEDASSOCIATION) {
	COND_PopCondition(FALSE);	/* pop DIMSE abort */
	COND_PopCondition(FALSE);	/* pop DUL abort */
	if (verbose)
	    printf("Association Aborted\n");
    } else {
	errmsg("DIMSE Failure (aborting association):\n");
	COND_DumpConditions();
    /* some kind of error so abort the association */
	cond = ASC_abortAssociation(assoc);
    }

    cond = ASC_dropAssociation(assoc);
    if (!SUCCESS(cond)) {
	fprintf(stderr, "Cannot Drop Association:\n");
	COND_DumpConditions();
    }
    cond = ASC_destroyAssociation(&assoc);
    if (!SUCCESS(cond)) {
	fprintf(stderr, "Cannot Destroy Association:\n");
	COND_DumpConditions();
    }

#ifndef NODISPLAY
    /* inform display that the association has ended */
    DSU_stopAssociation(DSU_MainAssociation);
#endif

    return cond;
}


/*
** CVS Log
** $Log: sce.cc,v $
** Revision 1.3  1996/09/27 08:54:25  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C
**
** Revision 1.2  1996/08/29 16:16:06  hewett
** Corrected copyright header.
**
** Revision 1.1.1.1  1996/04/26 14:16:01  hewett
** Initial Import.
**
**
*/
